@@ -1,18 +1,18 @@
 /*
    rdesktop: A Remote Desktop Protocol client.
    Miscellaneous protocol constants
-   Copyright (C) Matthew Chapman 1999-2002
+   Copyright (C) Matthew Chapman 1999-2005
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
@@ -117,6 +117,7 @@ enum RDP_DATA_PDU_TYPE
 	RDP_DATA_PDU_BELL = 34,
 	RDP_DATA_PDU_LOGON = 38,
 	RDP_DATA_PDU_FONT2 = 39,
+	RDP_DATA_PDU_KEYBOARD_INDICATORS = 41,
 	RDP_DATA_PDU_DISCONNECT = 47
 };
 
@@ -199,6 +200,15 @@ enum RDP_INPUT_DEVICE
 #define TEXT2_VERTICAL		0x04
 #define TEXT2_IMPLICIT_X	0x20
 
+/* RDP bitmap cache (version 2) constants */
+#define BMPCACHE2_C0_CELLS	0x78
+#define BMPCACHE2_C1_CELLS	0x78
+#define BMPCACHE2_C2_CELLS	0x150
+#define BMPCACHE2_NUM_PSTCELLS	0x9f6
+
+#define PDU_FLAG_FIRST		0x01
+#define PDU_FLAG_LAST		0x02
+
 /* RDP capabilities */
 #define RDP_CAPSET_GENERAL	1	/* Maps to generalCapabilitySet in T.128 page 138 */
 #define RDP_CAPLEN_GENERAL	0x18
@@ -231,15 +241,32 @@ enum RDP_INPUT_DEVICE
 #define RDP_CAPSET_COLCACHE	10
 #define RDP_CAPLEN_COLCACHE	0x08
 
-#define RDP_CAPSET_UNKNOWN	13
-#define RDP_CAPLEN_UNKNOWN	0x9C
+#define RDP_CAPSET_BMPCACHE2	19
+#define RDP_CAPLEN_BMPCACHE2	0x28
+#define BMPCACHE2_FLAG_PERSIST	((uint32)1<<31)
 
 #define RDP_SOURCE		"MSTSC"
 
 /* Logon flags */
-#define RDP_LOGON_NORMAL	0x33
-#define RDP_LOGON_AUTO		0x8
-#define RDP_LOGON_BLOB		0x100
+#define RDP_LOGON_AUTO		0x0008
+#define RDP_LOGON_NORMAL	0x0033
+#define RDP_COMPRESSION		0x0080
+#define RDP_LOGON_BLOB		0x0100
+#define RDP_LOGON_LEAVE_AUDIO	0x2000
+
+#define RDP5_DISABLE_NOTHING	0x00
+#define RDP5_NO_WALLPAPER	0x01
+#define RDP5_NO_FULLWINDOWDRAG	0x02
+#define RDP5_NO_MENUANIMATIONS	0x04
+#define RDP5_NO_THEMING		0x08
+#define RDP5_NO_CURSOR_SHADOW	0x20
+#define RDP5_NO_CURSORSETTINGS	0x40	/* disables cursor blinking */
+
+/* compression types */
+#define RDP_MPPC_COMPRESSED	0x20
+#define RDP_MPPC_RESET		0x40
+#define RDP_MPPC_FLUSH		0x80
+#define RDP_MPPC_DICT_SIZE      8192
 
 /* Keymap flags */
 #define MapRightShiftMask   (1<<0)
@@ -314,6 +341,61 @@ enum RDP_INPUT_DEVICE
 
 /* NT status codes for RDPDR */
 #define STATUS_SUCCESS			0x00000000
+#define STATUS_NOT_IMPLEMENTED          0x00000001
+#define STATUS_PENDING                  0x00000103
+
+#define STATUS_NO_MORE_FILES            0x80000006
+#define STATUS_DEVICE_PAPER_EMPTY       0x8000000e
+#define STATUS_DEVICE_POWERED_OFF       0x8000000f
+#define STATUS_DEVICE_OFF_LINE          0x80000010
+#define STATUS_DEVICE_BUSY              0x80000011
+
+#define STATUS_INVALID_HANDLE           0xc0000008
 #define STATUS_INVALID_PARAMETER	0xc000000d
+#define STATUS_NO_SUCH_FILE             0xc000000f
 #define STATUS_INVALID_DEVICE_REQUEST	0xc0000010
 #define STATUS_ACCESS_DENIED		0xc0000022
+#define STATUS_OBJECT_NAME_COLLISION    0xc0000035
+#define STATUS_DISK_FULL                0xc000007f
+#define STATUS_FILE_IS_A_DIRECTORY      0xc00000ba
+#define STATUS_NOT_SUPPORTED            0xc00000bb
+#define STATUS_TIMEOUT                  0xc0000102
+#define STATUS_NOTIFY_ENUM_DIR          0xc000010c
+#define STATUS_CANCELLED                0xc0000120
+
+
+/* RDPDR constants */
+#define RDPDR_MAX_DEVICES               0x10
+#define DEVICE_TYPE_SERIAL              0x01
+#define DEVICE_TYPE_PARALLEL            0x02
+#define DEVICE_TYPE_PRINTER             0x04
+#define DEVICE_TYPE_DISK                0x08
+#define DEVICE_TYPE_SCARD               0x20
+
+#define FILE_DIRECTORY_FILE             0x00000001
+#define FILE_NON_DIRECTORY_FILE         0x00000040
+#define FILE_COMPLETE_IF_OPLOCKED       0x00000100
+#define FILE_DELETE_ON_CLOSE            0x00001000
+#define FILE_OPEN_FOR_FREE_SPACE_QUERY  0x00800000
+
+/* RDP5 disconnect PDU */
+#define exDiscReasonNoInfo				0x0000
+#define exDiscReasonAPIInitiatedDisconnect		0x0001
+#define exDiscReasonAPIInitiatedLogoff			0x0002
+#define exDiscReasonServerIdleTimeout			0x0003
+#define exDiscReasonServerLogonTimeout			0x0004
+#define exDiscReasonReplacedByOtherConnection		0x0005
+#define exDiscReasonOutOfMemory				0x0006
+#define exDiscReasonServerDeniedConnection		0x0007
+#define exDiscReasonServerDeniedConnectionFips		0x0008
+#define exDiscReasonLicenseInternal			0x0100
+#define exDiscReasonLicenseNoLicenseServer		0x0101
+#define exDiscReasonLicenseNoLicense			0x0102
+#define exDiscReasonLicenseErrClientMsg			0x0103
+#define exDiscReasonLicenseHwidDoesntMatchLicense	0x0104
+#define exDiscReasonLicenseErrClientLicense		0x0105
+#define exDiscReasonLicenseCantFinishProtocol		0x0106
+#define exDiscReasonLicenseClientEndedProtocol		0x0107
+#define exDiscReasonLicenseErrClientEncryption		0x0108
+#define exDiscReasonLicenseCantUpgradeLicense		0x0109
+#define exDiscReasonLicenseNoRemoteConnections		0x010a
