@@ -1,7 +1,7 @@
 /*
    rdesktop: A Remote Desktop Protocol client.
    RDP licensing negotiation
-   Copyright (C) Matthew Chapman 1999-2002
+   Copyright (C) Matthew Chapman 1999-2005
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,15 +19,10 @@
 */
 
 #include "rdesktop.h"
-
-#ifdef WITH_OPENSSL
 #include <openssl/rc4.h>
-#else
-#include "crypto/rc4.h"
-#endif
 
 extern char g_username[16];
-extern char hostname[16];
+extern char g_hostname[16];
 
 static uint8 g_licence_key[16];
 static uint8 g_licence_sign_key[16];
@@ -36,27 +31,27 @@ BOOL g_licence_issued = False;
 
 /* Generate a session key and RC4 keys, given client and server randoms */
 static void
-licence_generate_keys(uint8 * client_key, uint8 * server_key, uint8 * client_rsa)
+licence_generate_keys(uint8 * client_random, uint8 * server_random, uint8 * pre_master_secret)
 {
-	uint8 session_key[48];
-	uint8 temp_hash[48];
+	uint8 master_secret[48];
+	uint8 key_block[48];
 
-	/* Generate session key - two rounds of sec_hash_48 */
-	sec_hash_48(temp_hash, client_rsa, client_key, server_key, 65);
-	sec_hash_48(session_key, temp_hash, server_key, client_key, 65);
+	/* Generate master secret and then key material */
+	sec_hash_48(master_secret, pre_master_secret, client_random, server_random, 'A');
+	sec_hash_48(key_block, master_secret, server_random, client_random, 'A');
 
-	/* Store first 16 bytes of session key, for generating signatures */
-	memcpy(g_licence_sign_key, session_key, 16);
+	/* Store first 16 bytes of session key as MAC secret */
+	memcpy(g_licence_sign_key, key_block, 16);
 
-	/* Generate RC4 key */
-	sec_hash_16(g_licence_key, &session_key[16], client_key, server_key);
+	/* Generate RC4 key from next 16 bytes */
+	sec_hash_16(g_licence_key, &key_block[16], client_random, server_random);
 }
 
 static void
 licence_generate_hwid(uint8 * hwid)
 {
 	buf_out_uint32(hwid, 2);
-	strncpy((char *) (hwid + 4), hostname, LICENCE_HWID_SIZE - 4);
+	strncpy((char *) (hwid + 4), g_hostname, LICENCE_HWID_SIZE - 4);
 }
 
 /* Present an existing licence to the server */
@@ -174,7 +169,7 @@ licence_process_demand(STREAM s)
 		return;
 	}
 
-	licence_send_request(null_data, null_data, g_username, hostname);
+	licence_send_request(null_data, null_data, g_username, g_hostname);
 }
 
 /* Send an authentication response packet */
