@@ -1,9 +1,9 @@
 /* bitmap.c */
-BOOL bitmap_decompress(unsigned char *output, int width, int height, unsigned char *input, int size,
-		       int Bpp);
+BOOL bitmap_decompress(uint8 * output, int width, int height, uint8 * input, int size, int Bpp);
 /* cache.c */
 HBITMAP cache_get_bitmap(uint8 cache_id, uint16 cache_idx);
-void cache_put_bitmap(uint8 cache_id, uint16 cache_idx, HBITMAP bitmap);
+void cache_put_bitmap(uint8 cache_id, uint16 cache_idx, HBITMAP bitmap, uint32 stamp);
+void cache_save_state(void);
 FONTGLYPH *cache_get_font(uint8 font, uint16 character);
 void cache_put_font(uint8 font, uint16 character, uint16 offset, uint16 baseline, uint16 width,
 		    uint16 height, HGLYPH pixmap);
@@ -26,12 +26,22 @@ void cliprdr_send_native_format_announce(uint8 * data, uint32 length);
 void cliprdr_send_data_request(uint32 format);
 void cliprdr_send_data(uint8 * data, uint32 length);
 BOOL cliprdr_init(void);
+/* disk.c */
+int disk_enum_devices(uint32 * id, char *optarg);
+NTSTATUS disk_query_information(NTHANDLE handle, uint32 info_class, STREAM out);
+NTSTATUS disk_set_information(NTHANDLE handle, uint32 info_class, STREAM in, STREAM out);
+NTSTATUS disk_query_volume_information(NTHANDLE handle, uint32 info_class, STREAM out);
+NTSTATUS disk_query_directory(NTHANDLE handle, uint32 info_class, char *pattern, STREAM out);
+NTSTATUS disk_create_notify(NTHANDLE handle, uint32 info_class);
+NTSTATUS disk_check_notify(NTHANDLE handle);
+/* mppc.c */
+int mppc_expand(uint8 * data, uint32 clen, uint8 ctype, uint32 * roff, uint32 * rlen);
 /* ewmhints.c */
 int get_current_workarea(uint32 * x, uint32 * y, uint32 * width, uint32 * height);
 /* iso.c */
 STREAM iso_init(int length);
 void iso_send(STREAM s);
-STREAM iso_recv(void);
+STREAM iso_recv(uint8 * rdpver);
 BOOL iso_connect(char *server, char *username);
 void iso_disconnect(void);
 /* licence.c */
@@ -40,13 +50,26 @@ void licence_process(STREAM s);
 STREAM mcs_init(int length);
 void mcs_send_to_channel(STREAM s, uint16 channel);
 void mcs_send(STREAM s);
-STREAM mcs_recv(uint16 * channel);
+STREAM mcs_recv(uint16 * channel, uint8 * rdpver);
 BOOL mcs_connect(char *server, STREAM mcs_data, char *username);
 void mcs_disconnect(void);
 /* orders.c */
 void process_orders(STREAM s, uint16 num_orders);
 void reset_order_state(void);
+/* parallel.c */
+int parallel_enum_devices(uint32 * id, char *optarg);
 /* printer.c */
+int printer_enum_devices(uint32 * id, char *optarg);
+/* printercache.c */
+int printercache_load_blob(char *printer_name, uint8 ** data);
+void printercache_process(STREAM s);
+/* pstcache.c */
+void pstcache_touch_bitmap(uint8 id, uint16 idx, uint32 stamp);
+BOOL pstcache_load_bitmap(uint8 id, uint16 idx);
+BOOL pstcache_put_bitmap(uint8 id, uint16 idx, uint8 * bmp_id, uint16 wd,
+			 uint16 ht, uint16 len, uint8 * data);
+int pstcache_enumerate(uint8 id, uint8 * list);
+BOOL pstcache_init(uint8 id);
 /* rdesktop.c */
 int main(int argc, char *argv[]);
 void generate_random(uint8 * random);
@@ -56,13 +79,24 @@ void xfree(void *mem);
 void error(char *format, ...);
 void warning(char *format, ...);
 void unimpl(char *format, ...);
-void hexdump(unsigned char *p, int len);
+void hexdump(unsigned char *p, unsigned int len);
+char *next_arg(char *src, char needle);
+void toupper_str(char *p);
+char *l_to_a(long N, int base);
 int load_licence(unsigned char **data);
 void save_licence(unsigned char *data, int length);
+BOOL rd_pstcache_mkdir(void);
+int rd_open_file(char *filename);
+void rd_close_file(int fd);
+int rd_read_file(int fd, void *ptr, int len);
+int rd_write_file(int fd, void *ptr, int len);
+int rd_lseek_file(int fd, int offset);
+BOOL rd_lock_file(int fd, int start, int len);
 /* rdp5.c */
-void rdp5_process(STREAM s, BOOL encryption);
+void rdp5_process(STREAM s);
 /* rdp.c */
 void rdp_out_unistr(STREAM s, char *string, int len);
+int rdp_in_unistr(STREAM s, char *string, int uni_len);
 void rdp_send_input(uint32 time, uint16 message_type, uint16 device_flags, uint16 param1,
 		    uint16 param2);
 void process_colour_pointer_pdu(STREAM s);
@@ -70,24 +104,22 @@ void process_cached_pointer_pdu(STREAM s);
 void process_system_pointer_pdu(STREAM s);
 void process_bitmap_updates(STREAM s);
 void process_palette(STREAM s);
-BOOL rdp_main_loop(void);
+BOOL rdp_loop(BOOL * deactivated, uint32 * ext_disc_reason);
+void rdp_main_loop(BOOL * deactivated, uint32 * ext_disc_reason);
 BOOL rdp_connect(char *server, uint32 flags, char *domain, char *password, char *command,
 		 char *directory);
 void rdp_disconnect(void);
 /* rdpdr.c */
-void rdpdr_send_connect(void);
-void rdpdr_send_name(void);
-void rdpdr_send_available(void);
-void rdpdr_send_completion(uint32 device, uint32 id, uint32 status, uint32 result, uint8 * buffer,
-			   uint32 length);
+int get_device_index(NTHANDLE handle);
+void convert_to_unix_filename(char *filename);
 BOOL rdpdr_init(void);
+void rdpdr_add_fds(int *n, fd_set * rfds, fd_set * wfds, struct timeval *tv, BOOL * timeout);
+struct async_iorequest *rdpdr_remove_iorequest(struct async_iorequest *prev,
+					       struct async_iorequest *iorq);
+void rdpdr_check_fds(fd_set * rfds, fd_set * wfds, BOOL timed_out);
+BOOL rdpdr_abort_io(uint32 fd, uint32 major, NTSTATUS status);
 /* rdpsnd.c */
-STREAM rdpsnd_init_packet(uint16 type, uint16 size);
-void rdpsnd_send(STREAM s);
 void rdpsnd_send_completion(uint16 tick, uint8 packet_index);
-void rdpsnd_process_negotiate(STREAM in);
-void rdpsnd_process_unknown6(STREAM in);
-void rdpsnd_process(STREAM s);
 BOOL rdpsnd_init(void);
 /* rdpsnd_oss.c */
 BOOL wave_out_open(void);
@@ -108,16 +140,20 @@ STREAM sec_init(uint32 flags, int maxlen);
 void sec_send_to_channel(STREAM s, uint32 flags, uint16 channel);
 void sec_send(STREAM s, uint32 flags);
 void sec_process_mcs_data(STREAM s);
-STREAM sec_recv(void);
+STREAM sec_recv(uint8 * rdpver);
 BOOL sec_connect(char *server, char *username);
 void sec_disconnect(void);
 /* serial.c */
+int serial_enum_devices(uint32 * id, char *optarg);
+BOOL serial_get_timeout(NTHANDLE handle, uint32 length, uint32 * timeout, uint32 * itv_timeout);
+BOOL serial_get_event(NTHANDLE handle, uint32 * result);
 /* tcp.c */
 STREAM tcp_init(uint32 maxlen);
 void tcp_send(STREAM s);
 STREAM tcp_recv(STREAM s, uint32 length);
 BOOL tcp_connect(char *server);
 void tcp_disconnect(void);
+char *tcp_get_address(void);
 /* xclip.c */
 void ui_clip_format_announce(uint8 * data, uint32 length);
 void ui_clip_handle_data(uint8 * data, uint32 length);
@@ -138,10 +174,13 @@ uint16 ui_get_numlock_state(unsigned int state);
 void reset_modifier_keys(void);
 void rdp_send_scancode(uint32 time, uint16 flags, uint8 scancode);
 /* xwin.c */
+void ui_begin_update(void);
+void ui_end_update(void);
 BOOL get_key_state(unsigned int state, uint32 keysym);
 BOOL ui_init(void);
 void ui_deinit(void);
 BOOL ui_create_window(void);
+void ui_resize_window(void);
 void ui_destroy_window(void);
 void xwin_toggle_fullscreen(void);
 int ui_select(int rdp_socket);
