@@ -1,7 +1,7 @@
 /* -*- c-basic-offset: 8 -*-
    rdesktop: A Remote Desktop Protocol client.
    Entrypoint and utility functions
-   Copyright (C) Matthew Chapman 1999-2004
+   Copyright (C) Matthew Chapman 1999-2005
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@
 #include <sys/stat.h>		/* stat */
 #include <sys/time.h>		/* gettimeofday */
 #include <sys/times.h>		/* times */
+#include <ctype.h>		/* toupper */
 #include <errno.h>
 #include "rdesktop.h"
 
@@ -34,17 +35,13 @@
 #include <sys/un.h>		/* sockaddr_un */
 #endif
 
-#ifdef WITH_OPENSSL
 #include <openssl/md5.h>
-#else
-#include "crypto/md5.h"
-#endif
 
 char g_title[64] = "";
 char g_username[64];
-char hostname[16];
+char g_hostname[16];
 char keymapname[16];
-int keylayout = 0x409;		/* Defaults to US keyboard layout */
+int g_keylayout = 0x409;	/* Defaults to US keyboard layout */
 
 int g_width = 800;		/* width is special: If 0, the
 				   geometry will be fetched from
@@ -52,12 +49,16 @@ int g_width = 800;		/* width is special: If 0, the
 				   absolute value specifies the
 				   percent of the whole screen. */
 int g_height = 600;
-int tcp_port_rdp = TCP_PORT_RDP;
+int g_xpos = 0;
+int g_ypos = 0;
+extern int g_tcp_port_rdp;
 int g_server_bpp = 8;
 int g_win_button_size = 0;	/* If zero, disable single app mode */
 BOOL g_bitmap_compression = True;
 BOOL g_sendmotion = True;
-BOOL g_orders = True;
+BOOL g_bitmap_cache = True;
+BOOL g_bitmap_cache_persist_enable = False;
+BOOL g_bitmap_cache_precache = True;
 BOOL g_encryption = True;
 BOOL packet_encryption = True;
 BOOL g_desktop_save = True;
@@ -67,12 +68,20 @@ BOOL g_hide_decorations = False;
 BOOL g_use_rdp5 = True;
 BOOL g_console_session = False;
 BOOL g_numlock_sync = False;
-extern BOOL g_owncolmap;
+BOOL g_owncolmap = False;
+BOOL g_ownbackstore = True;	/* We can't rely on external BackingStore */
+uint32 g_embed_wnd;
+uint32 g_rdp5_performanceflags =
+	RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG | RDP5_NO_MENUANIMATIONS;
 
 #ifdef WITH_RDPSND
 BOOL g_rdpsnd = False;
 #endif
 
+extern RDPDR_DEVICE g_rdpdr_device[];
+extern uint32 g_num_devices;
+extern char *g_rdpdr_clientname;
+
 #ifdef RDP2VNC
 extern int rfb_port;
 extern int defer_time;
@@ -85,7 +94,7 @@ static void
 usage(char *program)
 {
 	fprintf(stderr, "rdesktop: A Remote Desktop Protocol client.\n");
-	fprintf(stderr, "Version " VERSION ". Copyright (C) 1999-2003 Matt Chapman.\n");
+	fprintf(stderr, "Version " VERSION ". Copyright (C) 1999-2005 Matt Chapman.\n");
 	fprintf(stderr, "See http://www.rdesktop.org/ for more information.\n\n");
 
 	fprintf(stderr, "Usage: %s [options] server[:port]\n", program);
@@ -103,6 +112,7 @@ usage(char *program)
 	fprintf(stderr, "   -g: desktop geometry (WxH)\n");
 	fprintf(stderr, "   -f: full-screen mode\n");
 	fprintf(stderr, "   -b: force bitmap updates\n");
+	fprintf(stderr, "   -B: use BackingStore of X-server (if available)\n");
 	fprintf(stderr, "   -e: disable encryption (French TS)\n");
 	fprintf(stderr, "   -E: disable encryption from client to server\n");
 	fprintf(stderr, "   -m: do not send motion events\n");
@@ -111,14 +121,134 @@ usage(char *program)
 	fprintf(stderr, "   -K: keep window manager key bindings\n");
 	fprintf(stderr, "   -S: caption button size (single application mode)\n");
 	fprintf(stderr, "   -T: window title\n");
-	fprintf(stderr, "   -N: enable numlock synchronisation\n");
+	fprintf(stderr, "   -N: enable numlock syncronization\n");
+	fprintf(stderr, "   -X: embed into another window with a given id.\n");
 	fprintf(stderr, "   -a: connection colour depth\n");
-	fprintf(stderr, "   -r: enable specified device redirection (currently: sound)\n");
+	fprintf(stderr, "   -z: enable rdp compression\n");
+	fprintf(stderr, "   -x: RDP5 experience (m[odem 28.8], b[roadband], l[an] or hex nr.)\n");
+	fprintf(stderr, "   -P: use persistent bitmap caching\n");
+	fprintf(stderr, "   -r: enable specified device redirection (this flag can be repeated)\n");
+	fprintf(stderr,
+		"         '-r comport:COM1=/dev/ttyS0': enable serial redirection of /dev/ttyS0 to COM1\n");
+	fprintf(stderr, "             or      COM1=/dev/ttyS0,COM2=/dev/ttyS1\n");
+	fprintf(stderr,
+		"         '-r disk:floppy=/mnt/floppy': enable redirection of /mnt/floppy to 'floppy' share\n");
+	fprintf(stderr, "             or   'floppy=/mnt/floppy,cdrom=/mnt/cdrom'\n");
+	fprintf(stderr, "         '-r clientname=<client name>': Set the client name displayed\n");
+	fprintf(stderr, "             for redirected disks\n");
+	fprintf(stderr,
+		"         '-r lptport:LPT1=/dev/lp0': enable parallel redirection of /dev/lp0 to LPT1\n");
+	fprintf(stderr, "             or      LPT1=/dev/lp0,LPT2=/dev/lp1\n");
+	fprintf(stderr, "         '-r printer:mydeskjet': enable printer redirection\n");
+	fprintf(stderr,
+		"             or      mydeskjet=\"HP LaserJet IIIP\" to enter server driver as well\n");
+	fprintf(stderr, "         '-r sound:[local|off|remote]': enable sound redirection\n");
+	fprintf(stderr, "                     remote would leave sound on server\n");
 	fprintf(stderr, "   -0: attach to console\n");
 	fprintf(stderr, "   -4: use RDP version 4\n");
 	fprintf(stderr, "   -5: use RDP version 5 (default)\n");
 }
 
+void
+print_disconnect_reason(uint16 reason)
+{
+	char *text;
+
+	switch (reason)
+	{
+		case exDiscReasonNoInfo:
+			text = "No information available";
+			break;
+
+		case exDiscReasonAPIInitiatedDisconnect:
+			text = "Server initiated disconnect";
+			break;
+
+		case exDiscReasonAPIInitiatedLogoff:
+			text = "Server initiated logoff";
+			break;
+
+		case exDiscReasonServerIdleTimeout:
+			text = "Server idle timeout reached";
+			break;
+
+		case exDiscReasonServerLogonTimeout:
+			text = "Server logon timeout reached";
+			break;
+
+		case exDiscReasonReplacedByOtherConnection:
+			text = "The session was replaced";
+			break;
+
+		case exDiscReasonOutOfMemory:
+			text = "The server is out of memory";
+			break;
+
+		case exDiscReasonServerDeniedConnection:
+			text = "The server denied the connection";
+			break;
+
+		case exDiscReasonServerDeniedConnectionFips:
+			text = "The server denied the connection for security reason";
+			break;
+
+		case exDiscReasonLicenseInternal:
+			text = "Internal licensing error";
+			break;
+
+		case exDiscReasonLicenseNoLicenseServer:
+			text = "No license server available";
+			break;
+
+		case exDiscReasonLicenseNoLicense:
+			text = "No valid license available";
+			break;
+
+		case exDiscReasonLicenseErrClientMsg:
+			text = "Invalid licensing message";
+			break;
+
+		case exDiscReasonLicenseHwidDoesntMatchLicense:
+			text = "Hardware id doesn't match software license";
+			break;
+
+		case exDiscReasonLicenseErrClientLicense:
+			text = "Client license error";
+			break;
+
+		case exDiscReasonLicenseCantFinishProtocol:
+			text = "Network error during licensing protocol";
+			break;
+
+		case exDiscReasonLicenseClientEndedProtocol:
+			text = "Licensing protocol was not completed";
+			break;
+
+		case exDiscReasonLicenseErrClientEncryption:
+			text = "Incorrect client license enryption";
+			break;
+
+		case exDiscReasonLicenseCantUpgradeLicense:
+			text = "Can't upgrade license";
+			break;
+
+		case exDiscReasonLicenseNoRemoteConnections:
+			text = "The server is not licensed to accept remote connections";
+			break;
+
+		default:
+			if (reason > 0x1000 && reason < 0x7fff)
+			{
+				text = "Internal protocol error";
+			}
+			else
+			{
+				text = "Unknown reason";
+			}
+	}
+	fprintf(stderr, "disconnect: %s.\n", text);
+}
+
 static BOOL
 read_password(char *password, int size)
 {
@@ -176,7 +306,7 @@ parse_server_and_port(char *server)
 		if (*server == '[' && p != NULL)
 		{
 			if (*(p + 1) == ':' && *(p + 2) != '\0')
-				tcp_port_rdp = strtol(p + 2, NULL, 10);
+				g_tcp_port_rdp = strtol(p + 2, NULL, 10);
 			/* remove the port number and brackets from the address */
 			*p = '\0';
 			strncpy(server, server + 1, strlen(server));
@@ -188,7 +318,7 @@ parse_server_and_port(char *server)
 		p = strchr(server, ':');
 		if (p != NULL)
 		{
-			tcp_port_rdp = strtol(p + 1, NULL, 10);
+			g_tcp_port_rdp = strtol(p + 1, NULL, 10);
 			*p = 0;
 		}
 	}
@@ -196,7 +326,7 @@ parse_server_and_port(char *server)
 	p = strchr(server, ':');
 	if (p != NULL)
 	{
-		tcp_port_rdp = strtol(p + 1, NULL, 10);
+		g_tcp_port_rdp = strtol(p + 1, NULL, 10);
 		*p = 0;
 	}
 #endif /* IPv6 */
@@ -213,17 +343,21 @@ main(int argc, char *argv[])
 	char password[64];
 	char shell[128];
 	char directory[32];
-	BOOL prompt_password, rdp_retval = False;
+	BOOL prompt_password, deactivated;
 	struct passwd *pw;
-	uint32 flags;
+	uint32 flags, ext_disc_reason = 0;
 	char *p;
 	int c;
+
 	int username_option = 0;
 
 	flags = RDP_LOGON_NORMAL;
 	prompt_password = False;
 	domain[0] = password[0] = shell[0] = directory[0] = 0;
 	strcpy(keymapname, "en-us");
+	g_embed_wnd = 0;
+
+	g_num_devices = 0;
 
 #ifdef RDP2VNC
 #define VNCOPT "V:Q:"
@@ -231,7 +365,8 @@ main(int argc, char *argv[])
 #define VNCOPT
 #endif
 
-	while ((c = getopt(argc, argv, VNCOPT "u:d:s:c:p:n:k:g:fbeEmCDKS:T:Na:r:045h?")) != -1)
+	while ((c = getopt(argc, argv,
+			   VNCOPT "u:d:s:c:p:n:k:g:fbBeEmzCDKS:T:NX:a:x:Pr:045h?")) != -1)
 	{
 		switch (c)
 		{
@@ -283,7 +418,7 @@ main(int argc, char *argv[])
 				break;
 
 			case 'n':
-				STRNCPY(hostname, optarg, sizeof(hostname));
+				STRNCPY(g_hostname, optarg, sizeof(g_hostname));
 				break;
 
 			case 'k':
@@ -306,7 +441,7 @@ main(int argc, char *argv[])
 				}
 
 				if (*p == 'x')
-					g_height = strtol(p + 1, NULL, 10);
+					g_height = strtol(p + 1, &p, 10);
 
 				if (g_height <= 0)
 				{
@@ -315,7 +450,16 @@ main(int argc, char *argv[])
 				}
 
 				if (*p == '%')
+				{
 					g_width = -g_width;
+					p++;
+				}
+
+				if (*p == '+' || *p == '-')
+					g_xpos = strtol(p, &p, 10);
+
+				if (*p == '+' || *p == '-')
+					g_ypos = strtol(p, NULL, 10);
 
 				break;
 
@@ -324,7 +468,11 @@ main(int argc, char *argv[])
 				break;
 
 			case 'b':
-				g_orders = False;
+				g_bitmap_cache = False;
+				break;
+
+			case 'B':
+				g_ownbackstore = False;
 				break;
 
 			case 'e':
@@ -374,6 +522,10 @@ main(int argc, char *argv[])
 				g_numlock_sync = True;
 				break;
 
+			case 'X':
+				g_embed_wnd = strtol(optarg, NULL, 10);
+				break;
+
 			case 'a':
 				g_server_bpp = strtol(optarg, NULL, 10);
 				if (g_server_bpp != 8 && g_server_bpp != 16 && g_server_bpp != 15
@@ -384,13 +536,102 @@ main(int argc, char *argv[])
 				}
 				break;
 
+			case 'z':
+				DEBUG(("rdp compression enabled\n"));
+				flags |= RDP_COMPRESSION;
+				break;
+
+			case 'x':
+				if (strncmp("modem", optarg, 1) == 0)
+				{
+					g_rdp5_performanceflags =
+						RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG |
+						RDP5_NO_MENUANIMATIONS | RDP5_NO_THEMING;
+				}
+				else if (strncmp("broadband", optarg, 1) == 0)
+				{
+					g_rdp5_performanceflags = RDP5_NO_WALLPAPER;
+				}
+				else if (strncmp("lan", optarg, 1) == 0)
+				{
+					g_rdp5_performanceflags = RDP5_DISABLE_NOTHING;
+				}
+				else
+				{
+					g_rdp5_performanceflags = strtol(optarg, NULL, 16);
+				}
+				break;
+
+			case 'P':
+				g_bitmap_cache_persist_enable = True;
+				break;
+
 			case 'r':
-				if (!strcmp(optarg, "sound"))
+
+				if (strncmp("sound", optarg, 5) == 0)
+				{
+					optarg += 5;
+
+					if (*optarg == ':')
+					{
+						*optarg++;
+						while ((p = next_arg(optarg, ',')))
+						{
+							if (strncmp("remote", optarg, 6) == 0)
+								flags |= RDP_LOGON_LEAVE_AUDIO;
+
+							if (strncmp("local", optarg, 5) == 0)
 #ifdef WITH_RDPSND
-					g_rdpsnd = True;
+								g_rdpsnd = True;
 #else
-					warning("Not compiled with sound support");
+								warning("Not compiled with sound support\n");
 #endif
+
+							if (strncmp("off", optarg, 3) == 0)
+#ifdef WITH_RDPSND
+								g_rdpsnd = False;
+#else
+								warning("Not compiled with sound support\n");
+#endif
+
+							optarg = p;
+						}
+					}
+					else
+					{
+#ifdef WITH_RDPSND
+						g_rdpsnd = True;
+#else
+						warning("Not compiled with sound support\n");
+#endif
+					}
+				}
+				else if (strncmp("disk", optarg, 4) == 0)
+				{
+					/* -r disk:h:=/mnt/floppy */
+					disk_enum_devices(&g_num_devices, optarg + 4);
+				}
+				else if (strncmp("comport", optarg, 7) == 0)
+				{
+					serial_enum_devices(&g_num_devices, optarg + 7);
+				}
+				else if (strncmp("lptport", optarg, 7) == 0)
+				{
+					parallel_enum_devices(&g_num_devices, optarg + 7);
+				}
+				else if (strncmp("printer", optarg, 7) == 0)
+				{
+					printer_enum_devices(&g_num_devices, optarg + 7);
+				}
+				else if (strncmp("clientname", optarg, 7) == 0)
+				{
+					g_rdpdr_clientname = xmalloc(strlen(optarg + 11) + 1);
+					strcpy(g_rdpdr_clientname, optarg + 11);
+				}
+				else
+				{
+					warning("Unknown -r argument\n\n\tPossible arguments are: comport, disk, lptport, printer, sound\n");
+				}
 				break;
 
 			case '0':
@@ -413,7 +654,7 @@ main(int argc, char *argv[])
 		}
 	}
 
-	if (argc - optind < 1)
+	if (argc - optind != 1)
 	{
 		usage(argv[0]);
 		return 1;
@@ -434,7 +675,7 @@ main(int argc, char *argv[])
 		STRNCPY(g_username, pw->pw_name, sizeof(g_username));
 	}
 
-	if (hostname[0] == 0)
+	if (g_hostname[0] == 0)
 	{
 		if (gethostname(fullhostname, sizeof(fullhostname)) == -1)
 		{
@@ -446,7 +687,13 @@ main(int argc, char *argv[])
 		if (p != NULL)
 			*p = 0;
 
-		STRNCPY(hostname, fullhostname, sizeof(hostname));
+		STRNCPY(g_hostname, fullhostname, sizeof(g_hostname));
+	}
+
+	if ((flags & RDP_COMPRESSION) && (g_server_bpp > 8))
+	{
+		warning("rdp compression not supported for bpp > 8, compression disabled\n");
+		flags ^= RDP_COMPRESSION;
 	}
 
 	if (prompt_password && read_password(password, sizeof(password)))
@@ -470,7 +717,7 @@ main(int argc, char *argv[])
 	if (g_rdpsnd)
 		rdpsnd_init();
 #endif
-	/* rdpdr_init(); */
+	rdpdr_init();
 
 	if (!rdp_connect(server, flags, domain, password, shell, directory))
 		return 1;
@@ -486,18 +733,37 @@ main(int argc, char *argv[])
 
 	if (ui_create_window())
 	{
-		rdp_retval = rdp_main_loop();
+		rdp_main_loop(&deactivated, &ext_disc_reason);
 		ui_destroy_window();
 	}
 
 	DEBUG(("Disconnecting...\n"));
 	rdp_disconnect();
+	cache_save_state();
 	ui_deinit();
 
-	if (True == rdp_retval)
+	if (ext_disc_reason >= 2)
+		print_disconnect_reason(ext_disc_reason);
+
+	if (deactivated)
+	{
+		/* clean disconnect */
 		return 0;
+	}
 	else
-		return 2;
+	{
+		if (ext_disc_reason == exDiscReasonAPIInitiatedDisconnect
+		    || ext_disc_reason == exDiscReasonAPIInitiatedLogoff)
+		{
+			/* not so clean disconnect, but nothing to worry about */
+			return 0;
+		}
+		else
+		{
+			/* return error */
+			return 2;
+		}
+	}
 
 #endif
 
@@ -603,7 +869,11 @@ xmalloc(int size)
 void *
 xrealloc(void *oldmem, int size)
 {
-	void *mem = realloc(oldmem, size);
+	void *mem;
+
+	if (size < 1)
+		size = 1;
+	mem = realloc(oldmem, size);
 	if (mem == NULL)
 	{
 		error("xrealloc %d\n", size);
@@ -660,7 +930,7 @@ unimpl(char *format, ...)
 
 /* produce a hex dump */
 void
-hexdump(unsigned char *p, int len)
+hexdump(unsigned char *p, unsigned int len)
 {
 	unsigned char *line = p;
 	int i, thisline, offset = 0;
@@ -687,6 +957,123 @@ hexdump(unsigned char *p, int len)
 	}
 }
 
+/*
+  input: src is the string we look in for needle.
+  	 Needle may be escaped by a backslash, in
+	 that case we ignore that particular needle.
+  return value: returns next src pointer, for
+  	succesive executions, like in a while loop
+	if retval is 0, then there are no more args.
+  pitfalls:
+  	src is modified. 0x00 chars are inserted to
+	terminate strings.
+	return val, points on the next val chr after ins
+	0x00
+
+	example usage:
+	while( (pos = next_arg( optarg, ',')) ){
+		printf("%s\n",optarg);
+		optarg=pos;
+	}
+
+*/
+char *
+next_arg(char *src, char needle)
+{
+	char *nextval;
+	char *p;
+	char *mvp = 0;
+
+	/* EOS */
+	if (*src == (char) 0x00)
+		return 0;
+
+	p = src;
+	/*  skip escaped needles */
+	while ((nextval = strchr(p, needle)))
+	{
+		mvp = nextval - 1;
+		/* found backslashed needle */
+		if (*mvp == '\\' && (mvp > src))
+		{
+			/* move string one to the left */
+			while (*(mvp + 1) != (char) 0x00)
+			{
+				*mvp = *(mvp + 1);
+				*mvp++;
+			}
+			*mvp = (char) 0x00;
+			p = nextval;
+		}
+		else
+		{
+			p = nextval + 1;
+			break;
+		}
+
+	}
+
+	/* more args available */
+	if (nextval)
+	{
+		*nextval = (char) 0x00;
+		return ++nextval;
+	}
+
+	/* no more args after this, jump to EOS */
+	nextval = src + strlen(src);
+	return nextval;
+}
+
+
+void
+toupper_str(char *p)
+{
+	while (*p)
+	{
+		if ((*p >= 'a') && (*p <= 'z'))
+			*p = toupper((int) *p);
+		p++;
+	}
+}
+
+
+/* not all clibs got ltoa */
+#define LTOA_BUFSIZE (sizeof(long) * 8 + 1)
+
+char *
+l_to_a(long N, int base)
+{
+	static char ret[LTOA_BUFSIZE];
+
+	char *head = ret, buf[LTOA_BUFSIZE], *tail = buf + sizeof(buf);
+
+	register int divrem;
+
+	if (base < 36 || 2 > base)
+		base = 10;
+
+	if (N < 0)
+	{
+		*head++ = '-';
+		N = -N;
+	}
+
+	tail = buf + sizeof(buf);
+	*--tail = 0;
+
+	do
+	{
+		divrem = N % base;
+		*--tail = (divrem <= 9) ? divrem + '0' : divrem + 'a' - 10;
+		N /= base;
+	}
+	while (N);
+
+	strcpy(head, tail);
+	return ret;
+}
+
 
 int
 load_licence(unsigned char **data)
@@ -699,8 +1086,8 @@ load_licence(unsigned char **data)
 	if (home == NULL)
 		return -1;
 
-	path = (char *) xmalloc(strlen(home) + strlen(hostname) + sizeof("/.rdesktop/licence."));
-	sprintf(path, "%s/.rdesktop/licence.%s", home, hostname);
+	path = (char *) xmalloc(strlen(home) + strlen(g_hostname) + sizeof("/.rdesktop/licence."));
+	sprintf(path, "%s/.rdesktop/licence.%s", home, g_hostname);
 
 	fd = open(path, O_RDONLY);
 	if (fd == -1)
@@ -726,7 +1113,7 @@ save_licence(unsigned char *data, int length)
 	if (home == NULL)
 		return;
 
-	path = (char *) xmalloc(strlen(home) + strlen(hostname) + sizeof("/.rdesktop/licence."));
+	path = (char *) xmalloc(strlen(home) + strlen(g_hostname) + sizeof("/.rdesktop/licence."));
 
 	sprintf(path, "%s/.rdesktop", home);
 	if ((mkdir(path, 0700) == -1) && errno != EEXIST)
@@ -737,7 +1124,7 @@ save_licence(unsigned char *data, int length)
 
 	/* write licence to licence.hostname.new, then atomically rename to licence.hostname */
 
-	sprintf(path, "%s/.rdesktop/licence.%s", home, hostname);
+	sprintf(path, "%s/.rdesktop/licence.%s", home, g_hostname);
 	tmppath = (char *) xmalloc(strlen(path) + sizeof(".new"));
 	strcpy(tmppath, path);
 	strcat(tmppath, ".new");
@@ -764,3 +1151,95 @@ save_licence(unsigned char *data, int length)
 	xfree(tmppath);
 	xfree(path);
 }
+
+/* Create the bitmap cache directory */
+BOOL
+rd_pstcache_mkdir(void)
+{
+	char *home;
+	char bmpcache_dir[256];
+
+	home = getenv("HOME");
+
+	if (home == NULL)
+		return False;
+
+	sprintf(bmpcache_dir, "%s/%s", home, ".rdesktop");
+
+	if ((mkdir(bmpcache_dir, S_IRWXU) == -1) && errno != EEXIST)
+	{
+		perror(bmpcache_dir);
+		return False;
+	}
+
+	sprintf(bmpcache_dir, "%s/%s", home, ".rdesktop/cache");
+
+	if ((mkdir(bmpcache_dir, S_IRWXU) == -1) && errno != EEXIST)
+	{
+		perror(bmpcache_dir);
+		return False;
+	}
+
+	return True;
+}
+
+/* open a file in the .rdesktop directory */
+int
+rd_open_file(char *filename)
+{
+	char *home;
+	char fn[256];
+	int fd;
+
+	home = getenv("HOME");
+	if (home == NULL)
+		return -1;
+	sprintf(fn, "%s/.rdesktop/%s", home, filename);
+	fd = open(fn, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
+	if (fd == -1)
+		perror(fn);
+	return fd;
+}
+
+/* close file */
+void
+rd_close_file(int fd)
+{
+	close(fd);
+}
+
+/* read from file*/
+int
+rd_read_file(int fd, void *ptr, int len)
+{
+	return read(fd, ptr, len);
+}
+
+/* write to file */
+int
+rd_write_file(int fd, void *ptr, int len)
+{
+	return write(fd, ptr, len);
+}
+
+/* move file pointer */
+int
+rd_lseek_file(int fd, int offset)
+{
+	return lseek(fd, offset, SEEK_SET);
+}
+
+/* do a write lock on a file */
+BOOL
+rd_lock_file(int fd, int start, int len)
+{
+	struct flock lock;
+
+	lock.l_type = F_WRLCK;
+	lock.l_whence = SEEK_SET;
+	lock.l_start = start;
+	lock.l_len = len;
+	if (fcntl(fd, F_SETLK, &lock) == -1)
+		return False;
+	return True;
+}
